/*************************************************************/
/*   An Introduction to LLVS --- Rabi Dutta    Oct,  1989    */
/*                               Kamal Souccar Sept, 1991    */
/*************************************************************/

#include <stdio.h>		/* standard I/O defs */
#include <math.h>		/* math defs */
#include <ctype.h>		/* character defs */

#include "llvs_per_plane.h" /* LLVS global declarations */

#ifdef NO_PPIX_DRIVER
/* scratch space */
#define LLVS_SCRATCH_AREA_SIZE 2048
long int llvs_scratch[LLVS_SCRATCH_AREA_SIZE];
#endif NO_PPIX_DRIVER

#define ALPHA      1
#define IMAGESIZE 64
#define SQR(x)    ((x)*(x))
#define MAX(x,y)  (((x)>(y)) ? (x) : (y))
#define DELTA      2
#define THRESHOLD  0
#define TRUE       1
#define FALSE      0

long int delx[3][3] = { { -1, 0, 1 },
			{ -2, 0, 2 },
			{ -1, 0, 1 } };

long int dely[3][3] = { { -1, -2, -1 },
			{  0,  0,  0 },
			{  1,  2,  1 } };

double conf_inc[4][4] = { { -DELTA,     0, -DELTA, -DELTA },
			  {      0, DELTA,  DELTA,  DELTA },
			  { -DELTA, DELTA,      0,      0 },
			  { -DELTA, DELTA,      0,      0 } };

main(argc, argv)
int argc;
char **argv;
{
  PLANE *plane1;
  PLANE_INFO *plane_info1;
  char *associations1;
  char filename[128];
  
  long int i,j,converged=FALSE;
  long int g[IMAGESIZE][IMAGESIZE], h[IMAGESIZE][IMAGESIZE];
  long int hx[IMAGESIZE][IMAGESIZE], hy[IMAGESIZE][IMAGESIZE];
  
  if(argc >1) strcpy(filename, argv[1]);
  else {
    printf("Enter filename: ");
    scanf("%s", filename);
  }
  read_plane(&plane1,&plane_info1,&associations1,filename);
  printf("read_plane successful\n");
  
  for (i=0;i<IMAGESIZE;++i)
    for (j=0;j<IMAGESIZE;++j)
      GET_PIXEL(g[i][j],9999,plane1,i,j,(*plane_info1));

  convolve(delx,g,hx);
  convolve(dely,g,hy);
  
  for (i=0;i<IMAGESIZE;++i)
    for (j=0;j<IMAGESIZE;++j)
      h[i][j] = (long int) sqrt( (double) (SQR(hx[i][j]) + SQR(hy[i][j])) );

/*  convolve(laplace,g,l); */
/*  zero_cross(l,h,ans);   */

  normalize(h); 

  for (i=0;i<IMAGESIZE;++i)
    for (j=0;j<IMAGESIZE;++j)
      SET_PIXEL(h[i][j],plane1,i,j,(*plane_info1));
  write_plane(plane1,plane_info1,associations1,"grad.plane");
  /* writes the new image into the file new.plane */

  while(!converged) {
    converged = relax(h,hx,hy);
    printf("."); fflush(stdout);
  }
  printf("\n");

  for (i=0;i<IMAGESIZE;++i)
    for (j=0;j<IMAGESIZE;++j)
      SET_PIXEL(h[i][j],plane1,i,j,(*plane_info1));
  write_plane(plane1,plane_info1,associations1,"relax.plane");
}

convolve(f,g,h)
long int f[2*ALPHA+1][2*ALPHA+1],g[IMAGESIZE][IMAGESIZE];
long int h[IMAGESIZE][IMAGESIZE];
{
  int x,y,i,j;

  for (x=ALPHA;x<(IMAGESIZE-ALPHA); ++x) {
    for (y=ALPHA;y<(IMAGESIZE-ALPHA); ++y) {
      h[x][y] = 0.0;
      for (i=(-ALPHA); i<=ALPHA; ++i)
	for (j=(-ALPHA); j<=ALPHA; ++j)
	  h[x][y] += f[i+ALPHA][j+ALPHA]*g[x+i][y+j];
    }
  }
}

normalize(image)
long int image[IMAGESIZE][IMAGESIZE];
{
  int i,j, max=0;
  
  for (i=0;i<IMAGESIZE;++i)
    for (j=0;j<IMAGESIZE;++j)
      if (abs(image[i][j]) > max) max = abs(image[i][j]);

  for (i=0;i<IMAGESIZE;++i)
    for (j=0;j<IMAGESIZE;++j)
      image[i][j] = (abs(image[i][j])*255)/max;  
}

int relax(grad,gx,gy)
long int grad[IMAGESIZE][IMAGESIZE],gx[IMAGESIZE][IMAGESIZE];
long int gy[IMAGESIZE][IMAGESIZE];
{
  long int i,j,left_type,right_type,def_type(),flag;
  long int left[3], right[3], new_g, off=3;

  flag = TRUE;
  for (i=1;i<(IMAGESIZE-1);++i) {
    for (j=1;j<(IMAGESIZE-1);++j) {
      if (grad[i][j] > THRESHOLD) {
	if (abs(gx[i][j]) > abs(gy[i][j])){ /* use top-bottom neighborhood */
	  left[0] = grad[i-off][j+off];
	  left[1] = grad[i-off][j];
	  left[2] = grad[i-off][j-off];
	  right[0] = grad[i+off][j+off];
	  right[1] = grad[i+off][j];
	  right[2] = grad[i+off][j-off];
	}
	else {                        /* use left-right neighborhood */
	  left[0] = grad[i-off][j-off];
	  left[1] = grad[i][j-off];
	  left[2] = grad[i+off][j-off];
	  right[0] = grad[i-off][j+off];
	  right[1] = grad[i][j+off];
	  right[2] = grad[i+off][j+off];
	}
      
	left_type = def_type(left);
	right_type = def_type(right);

	/* printf(" %d %d type %d-%d inc = %6.4lf\n",
	       i,j,left_type, right_type, conf_inc[left_type][right_type]); */
	new_g = grad[i][j] + conf_inc[left_type][right_type];
	if (new_g > 255) new_g = 255;
	if (new_g < 0) new_g = 0;
	if (new_g != grad[i][j]) flag = FALSE;
	grad[i][j] = new_g;
      }
      else grad[i][j] = 0;
    }
  }
  return(flag);
}

long int def_type(neighbors)
long int neighbors[3];
{
  long int i,a,b,c,type;
  long int m, q=1, conf[4];

  if (neighbors[0] > neighbors[1]) {
    if (neighbors[0] > neighbors[2]) {
      a=0; 
      if (neighbors[1] > neighbors[2]) {    /* 0,1,2 */
	b=1; c=2;
      }
      else{                                 /* 0,2,1 */
	b=2; c=1;
      }
    }
    else { a=2; b=0; c=1; }                 /* 2,0,1 */
  }
  else {
    if (neighbors[1] > neighbors[2]){
      a=1;
      if (neighbors[0] > neighbors[2]) {    /* 1,0,2 */
	b=0; c=2;
      }
      else {                                /* 1,2,0 */
	b=2; c=0;
      }
    }
    else { a=2; b=1; c=0;}                  /* 2,1,0 */
  }

  m = MAX(q,neighbors[a]);
/*  printf("m=%6.4lf a=%6.4lf b=%6.4lf c=%6.4lf q=%6.4lf\n",
	 m,neighbors[a],neighbors[b],neighbors[c],q); */

  conf[0] = (m-neighbors[a])*(m-neighbors[b])*(m-neighbors[c]);
  conf[1] = neighbors[a]*(m-neighbors[b])*(m-neighbors[c]);
  conf[2] = neighbors[a]*neighbors[b]*(m-neighbors[c]);
  conf[3] = neighbors[a]*neighbors[b]*neighbors[c];

  type = 0;
  for (i=1;i<4;++i) {
    if (conf[i] > conf[type]) type = i;
  }
  return(type);
}
